#!/usr/bin/perl

# Script que gerencia o número de processos correntes, para o parser
# mantendo um número limite de processo em execução e chamando um novo
# processo à medida que os processos em execução finalizam.
#
# A idéia é usá-lo para fazer o parse de textos sentença por sentença,
# para que aquelas que demandam muito tempo de processamento não sejam
# um empecilho para a continuidade da análise. É claro que pode ocorrer
# de termos, por exemplo, 30-50 processos demorados rodando junto, mas
# ainda assim, o processo será mais eficiente do que para todo o arquivo
# de uma vez só.
#
# Uso: tb-backgroundParsing.pl <max. heap> <max_active_processes> <max_parse_time> <preparse-file>
#
# Funcionamento:
#   - Criação de pasta temporária, para o split do arquivo preparse
#   - Loop de carregamento de processos, verificando a lista de sentenças
#     a analisar (aquelas que não contém o arquivo ".parsed").
#   - Junção das partes novamente, para o arquivo de saída (output)
#   - Correções automáticas sobre o arquivo gerado e exclusão da pasta temporária para o parsing.
#
# Dependêcias:
#
#   - Biblioteca Time::SoFar (script)
#   - 'tcsh' (shell, para os scripts do parser) ('sudo apt-get tcsh' deve resolver)
#
# Autor: Pablo Faria (28/08/09)

use File::Basename;
use Time::SoFar qw( runtime runinterval figuretimes );
use POSIX ":sys_wait_h";

local $| = 1; # autoflush

# Verifica parâmetros
if(scalar(@ARGV) != 4) {
   print "\nScripts do Corpus Tycho Brahe, IEL/Unicamp (2009)\n";
   print "Autor: Pablo Faria (pablofaria\@gmail.com)\n\n";
   print "Uso:\n    tb-backgroundParsing.pl <max_heap> <max_active_processes>\n";
   print "                                  <max_parse_time> <pos-file>\n";
   print "Saída:\n    <pos_file>.psd\n\n";
   print "         <max_heap>               Tamanho da pilha (em MB) (até 500)\n";
   print "         <max_active_processes>   Quantidade máxima de análises simultâneas\n";
   print "         <max_parse_time>         Tempo máximo de espera na análise de uma\n";
   print "                                  sentença (minutos)\n";
   print "         <pos-file>               Arquivo POS a processar\n\n";
   exit;
}

$main_pid = $$; # PID of the script

$SIG{CHLD} = 'IGNORE';  # Automatically reap zoombie processes

# Parser information
$dbparser = "/home/tycho/dbparser/";
$parse = $dbparser."bin/parse";

# Background parsing parameters
$max_active_processes = int($ARGV[1]);
$max_parse_time = int($ARGV[2]);

# Save current path
$cur_path = (`pwd`);

# Parse input file name
($name, $path, $suffix) = fileparse($ARGV[3]);

$preparse_file = $name.$suffix.".preparse";

if ($path eq "./") { $path = (`pwd`); $path =~ s/\n//g; $path .= "/"; }

$pos_file = $path.$name.$suffix;

print "----------------------------\n";

# Go to temporary folder
chdir "/tmp/";

# Generate 'preparse' file
print "[ Geração de versão 'preparse' para análise ]\n";
(`tb-preparseFromPOS.pl $pos_file`);

# Make temporary folder
(`rm -rf $name.parsing/`);
(`mkdir $name.parsing`);

# Move preparse file to new folder
(`mv $path$preparse_file $name.parsing/`);

# Enter new folder
chdir "$name.parsing/";

# Replace hifens by underscores  (generate '.tmp' file)
print "[ Coloca underscores em etiquetas ]\n";
(`tb-underscoreSubTags.pl preparse under $preparse_file`);
(`mv $preparse_file.tmp $preparse_file`);

# Split preparse file into single files containing one sentence each
print "[ Partição do arquivo em sub-arquivos contendo uma sentença ]\n";
(`rm -f *_s*.txt`);
(`tb-splitPreparse.pl $preparse_file`);

@preparse_files = glob("*_s*.txt");

$num = scalar(@preparse_files);

print "[ Análise sintática do arquivo <$name> iniciada ]\n[ Total: $num sentenças ]\n";

# Report about parser progress
#(`echo "Iniciou!\nTotal: $num sentenças" | mailx -s "parser:status[$name]" youremail\@thedomain`);

# Remove old parsed files
(`rm -f *.parsed`);
@parsed_files = glob("*.parsed");  # None, at this time

# Process each file, until all have a "parsed" version
my @pids;
my $pid_time = {};
my $trials = {};

my $active_proc = 1;  # Fake value, just to enter the while loop
my $last_mail_time = time;

while (scalar(@preparse_files) > scalar(@parsed_files) or $active_proc > 0) {
    # Start parses (if any is still needed)
    if (scalar(@preparse_files) > scalar(@parsed_files)) {
        foreach $file (@preparse_files) {
            $active_proc = int(`ps -A | grep " parse" | wc -l`);
            if ($active_proc < $max_active_processes) {
                if (not -e $file.".log") {
                    print ".";
                    (`echo "Started $file" > $file.log`);
                    # Execução do parser (em background)
                    $cmd = "$parse $ARGV[0] $dbparser"."settings/classicalport.properties $dbparser"."train.obj.gz $file >> $file.log 2>>$file.log";
                    die "[: could not fork :]" unless defined(my $pid = fork);
                    unless ($pid) { # Child execs
                        exec $cmd;
                        die "[: exec of parse failed to ($file) :]";
                    }
                    push @pids, $pid; # Parent stores children's pids
                    while (int(`ps -fA | grep "$file" | wc -l`) < 4) {
                        sleep 1; # Wait until the parse is fully up
                    }
                    @{$pid_time->{$pid}} = (time, $file);
                    $trials->{$file} = 1;
                }
            }
            else { last; }
        }
    }

    # Check whether a parse is too old (more than max_parse_time)
    #   (if so (and if not succesful) kill it and generate a flat tree for the sentence)
    my $fg_kill = 0;
    my $now = time;
    foreach $pid (@pids) {
        my $pid_f = $pid_time->{$pid}[1];
        if ($pid_f eq "") { next; }
        my $tmp = int(`ps -fA | grep "$pid_f" | wc -l`);
        my $active = ($tmp >= 4);
        if ($active and ($now - $pid_time->{$pid}[0]) > ($max_parse_time*60)) {
            # Get (the only two) child processes IDs
            $c_pid  = (`ps h -o pid --ppid $pid`);  # Child of parent ("tsch parse")
            $c_pid  =~ s/[\n ]//g;
            $cc_pid = " ".(`ps h -o pid --ppid $c_pid`);  # Child of the child ("java")
            $cc_pid =~ s/[\n ]//g;
            # Kill parse process
            (`kill -9 $pid $c_pid $cc_pid`);
            # Messages and flat structure generation
            delete $pid_time->{$pid};
            print "#";
            (`echo "\n[: Gave up parsing after $max_parse_time minutes. Assigning a flat tree structure... :]\n\n" >> $pid_f.log`);
            (`tb-assignFlatTree.pl $pid_f`);
            sleep 2;  # Some time for the system to kill the processes
            $fg_kill = 1;
        } elsif (not $active) {
            # Stop partly running parse (or 'zoombie' processes)
            # Get (the only two) child processes IDs
            $c_pid  = (`ps h -o pid --ppid $pid`);  # Child of parent ("tsch parse")
            $c_pid  =~ s/[\n ]//g;
            $cc_pid = "";
            if ($c_pid ne "") {
                $cc_pid = " ".(`ps h -o pid --ppid $c_pid`);  # Child of the child ("java")
                $cc_pid  =~ s/[\n ]//g;
            }
            # Kill parse process
            (`kill -9 $pid $c_pid $cc_pid`);
            sleep 2;

            # The parser finished or crashed for some reason
            # Remove 'log' and (empty) 'parsed' file so the parser can try again
            if (-e $pid_f.".log" and int(`grep "has a bad format:" $pid_f.log | wc -l`) == 1) {
               # Abort parse: the POS file must be examined
               delete $pid_time->{$pid};
               print "!\nAnálise abortada! Erro: formato de arquivo inválido, ver log.\n"; #"[: Desistiu de $pid_f após 3 tentativas :]";
               (`echo "\n[: Gave up parsing. Bad file format error. :]\n\n" >> $pid_f.log`);
               #(`cat $pid_f.log | mailx -s "parser:status[$name]" youremail\@thedomain`);
               die;
            } elsif (-e $pid_f.".log" and int(`grep "Have a nice day!" $pid_f.log | wc -l`) == 1) {
                # Successful parse
                delete $pid_time->{$pid};
            } elsif (-e $pid_f.".log" and int(`grep "OutOfMemoryError" $pid_f.log | wc -l`) > 0) {
                # OutOfMemory (abandon it and go flat!)
                delete $pid_time->{$pid};
                print "!"; #"[: Desistiu de $pid_f por falta de memória :]";
                (`echo "\n[: Could not parse (OutOfMemoryError). Assigning a flat tree structure... :]\n\n" >> $pid_f.log`);
                (`tb-assignFlatTree.pl $pid_f`);
                sleep 2;  # Some time for the system to kill the processes
            } else {
                # Remove 'log' and (empty) 'parsed' files for a new try
                if ($trials->{$pid_f} < 3) {
                    (`rm -f $pid_f.log $pid_f.parsed`);
                    delete $pid_time->{$pid};
                    $trials->{$pid_f} += 1;
                    $fg_kill = 1;
                } else {
                    # Give up after 3 trials
                    delete $pid_time->{$pid};
                    print "*"; #"[: Desistiu de $pid_f após 3 tentativas :]";
                    (`echo "\n[: Gave up after 3 trials. Assigning a flat tree structure... :]\n\n" >> $pid_f.log`);
                    (`tb-assignFlatTree.pl $pid_f`);
                    sleep 2;  # Some time for the system to kill the processes
                }
            }
        }
    }

    if (not $fg_kill) {
        # Usually a parse take at least this time to create 'log' and 'parsed' files
        sleep 10;
    }

    @parsed_files = glob("*.parsed");
    $active_proc = int(`ps -A | grep " parse" | wc -l`);

    # Hourly report about parser progress
    if (time > ($last_mail_time + 3600)) {
        $last_mail_time = time;
        #(`ls -l *.parsed 2>>/dev/null | wc -l | mailx -s "parser:status[$name]" youremail\@thedomain`);
    }
}

print "\n[ Junção do resultado da análise ]\n";

# Join the parsed files together in one file
$psd_file = $name.$suffix.".parsed";
$log_file = $name.$suffix.".log";
open(ARQ, ">$psd_file") || die;
@ordered_files = sort @preparse_files;
foreach $file (@ordered_files) {
    (`cat $file.log >> $log_file`);
    $tmp = (`cat $file.parsed`);
    $tmp = $tmp."\n";
    $tmp =~ s/^\n+//;
    $data .= $tmp;
}
$data =~ s/\n+$//;
print ARQ $data;
close(ARQ);

# Fix post parse commom problems (generate '.psd' file)
print "[ Correção de problemas comuns nas análises ]\n";
(`tb-fixPostParse.pl $psd_file`);
(`cp $psd_file.psd $path$name$suffix.bkp.psd`);

# Retrieve subtags (generate '.tmp' file)
#print "[ Recuperação de sub-etiquetas ]\n";
#(`tb-retrieveSubTags.pl $pos_file $psd_file.psd`);

# Replace underscores by hifens  (generate '.tmp' file)
print "[ Corrigindo hifens em etiquetas ]\n";
(`tb-underscoreSubTags.pl psd hifen $psd_file.psd`);

# Move
(`mv $psd_file.psd.tmp $path$name$suffix.psd`);
(`mv $log_file $path$name$suffix.log`);

# Down one level
chdir "../";

# Delete temporary files
print "[ Exclusão de dados temporários ]\n";
(`rm -rf $name.parsing/`);

# Return to previous folder
chdir "$cur_path";

print "----------------------------\n";
print "Análise sintática finalizada!\n";
$msg = "Estrutura 'flat' atribuída para ".int(`grep "IP-XXX" $path$name$suffix.psd | wc -l`)." sentenças!";
print $msg."\n";
(`echo "\n$msg" >> $path$name$suffix.log`);

my $elapsed = runtime();
(`echo "\nTempo decorrido: $elapsed" >> $path$name$suffix.log`);
printf("Tempo decorrido: $elapsed\n");
print "----------------------------\n";

# Report about parser progress
#(`echo "Finalizou!" | mailx -s "parser:status[$name]" youremail\@thedomain`);
